<?php

namespace Cursed\Screen;

/**
 *
 * @property get int $lines
 * @property get int $columns
 */
class Screen
{

    private int $lines;

    private int $columns;

    private string $buffer = '';

    public bool $zeroIndexed = true;

    private bool $initialized = false;

    public function __construct()
    {
        $this->setupTerminal();
        $this->measureTerminal();
        set_exception_handler(function (\Throwable $t) {
            $this->shutdownTerminal();
            fwrite(STDERR, "\e[31m#### FATAL ERROR #####\n{$t}\e[0m\n");
            exit(0);
        });
    }

    public function __destruct()
    {
        $this->shutdownTerminal();
    }

    private function setupTerminal(): void
    {
        if ($this->initialized) return;
        $this
            ->write("\e[s")
            ->write("\e[?25h")
            ->write("\e[?1047h")
            ->write("\e[1;1H")
            ->write("\e[0m")
            ->write("\e[2J")
            ->flush();
        $this->initialized = true;
    }

    private function shutdownTerminal(): void
    {
        if (!$this->initialized) return;
        $this
            ->write("\e[?1047l")
            ->write("\e[u")
            ->flush();
        $this->initialized = false;
    }

    private function measureTerminal(): void
    {
        $this->lines = intval(exec("tput lines"));
        $this->columns = intval(exec("tput cols"));
    }

    public function __get(string $name): mixed
    {
        switch ($name) {
            case 'lines': return $this->lines;
            case 'columns': return $this->columns;
        }
    }

    public function setCursorPosition(int $column, int $line): self
    {
        if ($line < 0) {
            $line = $this->lines + $line;
        }
        $this->printf("\e[%d;%dH", $this->zeroIndexed ? $line + 1 : $line, $this->zeroIndexed ? $column + 1 : $column);
        return $this;
    }

    public function setCursorVisible(bool $visible): self
    {
        $this->write("\e[?25".($visible?"h":"l"));
        return $this;
    }

    public function write(string $data): self
    {
        $this->buffer .= $data;
        return $this;
    }

    public function printf(string $format, ...$args): self
    {
        $this->write(sprintf($format, ...$args));
        return $this;
    }

    public function writeAt(int $column, int $line, string $data): self
    {
        $this->setCursorPosition($column, $line);
        $this->write($data);
        return $this;
    }

    public function writeLinesAt(int $column, int $line, array $lines): self
    {
        for ($n = 0; $n < count($lines); $n++) {
            $this->writeAt($column, $line + ($line > 0 ? $n : -$n), $lines[$n]);
        }
        return $this;
    }

    public function flush(): self
    {
        fwrite(STDOUT, $this->buffer);
        $this->buffer = '';
        return $this;
    }

    public function draw(Renderable $renderable, array $options = []): self
    {
        $renderable->render($this, $options);
        return $this;
    }

    public function pstrlen(string $text): int
    {
        $text = preg_replace("/\e\[(.+?)m/", "", $text);
        return mb_strlen($text);
    }

}
